<?php

declare(strict_types=1);

/*
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

namespace Chill\ActivityBundle\Export\Filter\ACPFilters;

use Chill\ActivityBundle\Export\Declarations;
use Chill\MainBundle\Entity\Scope;
use Chill\MainBundle\Export\FilterInterface;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Doctrine\ORM\Query\Expr\Andx;
use Doctrine\ORM\QueryBuilder;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Symfony\Component\Form\FormBuilderInterface;
use function in_array;

class UserScopeFilter implements FilterInterface
{
    public function __construct(private readonly TranslatableStringHelper $translatableStringHelper)
    {
    }

    public function addRole(): ?string
    {
        return null;
    }

    public function alterQuery(QueryBuilder $qb, $data)
    {
        if (!in_array('actuser', $qb->getAllAliases(), true)) {
            $qb->join('activity.user', 'actuser');
        }

        $where = $qb->getDQLPart('where');

        $clause = $qb->expr()->in('actuser.mainScope', ':userscope');

        if ($where instanceof Andx) {
            $where->add($clause);
        } else {
            $where = $qb->expr()->andX($clause);
        }

        $qb->add('where', $where);
        $qb->setParameter('userscope', $data['accepted_userscope']);
    }

    public function applyOn(): string
    {
        return Declarations::ACTIVITY_ACP;
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $builder->add('accepted_userscope', EntityType::class, [
            'class' => Scope::class,
            'choice_label' => fn (Scope $s) => $this->translatableStringHelper->localize(
                $s->getName()
            ),
            'multiple' => true,
            'expanded' => true,
        ]);
    }

    public function describeAction($data, $format = 'string'): array
    {
        $scopes = [];

        foreach ($data['accepted_userscope'] as $s) {
            $scopes[] = $this->translatableStringHelper->localize(
                $s->getName()
            );
        }

        return ['Filtered activity by userscope: only %scopes%', [
            '%scopes%' => implode(', ', $scopes),
        ]];
    }

    public function getTitle(): string
    {
        return 'Filter activity by userscope';
    }
}
