<?php

declare(strict_types=1);

/*
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

namespace Chill\PersonBundle\Export\Aggregator\AccompanyingCourseAggregators;

use Chill\MainBundle\Export\AggregatorInterface;
use Chill\MainBundle\Form\Type\PickRollingDateType;
use Chill\MainBundle\Repository\ScopeRepositoryInterface;
use Chill\MainBundle\Service\RollingDate\RollingDate;
use Chill\MainBundle\Service\RollingDate\RollingDateConverterInterface;
use Chill\MainBundle\Templating\TranslatableStringHelperInterface;
use Chill\PersonBundle\Export\Declarations;
use Doctrine\ORM\QueryBuilder;
use LogicException;
use Symfony\Component\Form\FormBuilderInterface;

class ReferrerScopeAggregator implements AggregatorInterface
{
    private const SCOPE_KEY = 'acp_agg_refscope_user_history_ref_scope_name';

    public function __construct(private readonly ScopeRepositoryInterface $scopeRepository, private readonly TranslatableStringHelperInterface $translatableStringHelper, private readonly RollingDateConverterInterface $rollingDateConverter)
    {
    }

    public function addRole(): ?string
    {
        return null;
    }

    public function alterQuery(QueryBuilder $qb, $data)
    {
        $userHistory = 'acp_agg_refscope_user_history';
        $ref = 'acp_agg_refscope_user_history_ref';
        $scopeName = self::SCOPE_KEY;
        $dateCalc = 'acp_agg_refscope_user_history_date_calc';

        $qb
            ->leftJoin('acp.userHistories', $userHistory)
            ->leftJoin($userHistory . '.user', $ref)
            ->andWhere(
                $qb->expr()->orX(
                    $qb->expr()->isNull($userHistory),
                    $qb->expr()->andX(
                        $qb->expr()->lte($userHistory . '.startDate', ':' . $dateCalc),
                        $qb->expr()->orX(
                            $qb->expr()->isNull($userHistory . '.endDate'),
                            $qb->expr()->gt($userHistory . '.endDate', ':' . $dateCalc)
                        )
                    )
                )
            )
            ->setParameter(
                $dateCalc,
                $this->rollingDateConverter->convert($data['date_calc'])
            );

        // add groups
        $qb
            ->addSelect('IDENTITY(' . $ref . '.mainScope) AS ' . $scopeName)
            ->addGroupBy($scopeName);
    }

    public function applyOn()
    {
        return Declarations::ACP_TYPE;
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $builder->add('date_calc', PickRollingDateType::class, [
            'data' => new RollingDate(RollingDate::T_TODAY),
            'label' => 'export.aggregator.course.by_user_scope.Computation date for referrer',
            'required' => true,
        ]);
    }

    public function getLabels($key, array $values, $data)
    {
        return function ($value) {
            if ('_header' === $value) {
                return 'export.aggregator.course.by_user_scope.Referrer\'s scope';
            }

            if (null === $value || '' === $value) {
                return '';
            }

            $scope = $this->scopeRepository->find($value);

            if (null === $scope) {
                throw new LogicException('no scope found with this id: ' . $value);
            }

            return $this->translatableStringHelper->localize($scope->getName());
        };
    }

    public function getQueryKeys($data)
    {
        return [self::SCOPE_KEY];
    }

    public function getTitle()
    {
        return 'export.aggregator.course.by_user_scope.Group course by referrer\'s scope';
    }
}
