<?php

declare(strict_types=1);

/*
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

namespace Chill\PersonBundle\Export\Export;

use Chill\CustomFieldsBundle\CustomFields\CustomFieldChoice;
use Chill\CustomFieldsBundle\Entity\CustomField;
use Chill\CustomFieldsBundle\Service\CustomFieldProvider;
use Chill\MainBundle\Export\ExportElementValidatedInterface;
use Chill\MainBundle\Export\FormatterInterface;
use Chill\MainBundle\Export\GroupedExportInterface;
use Chill\MainBundle\Export\Helper\ExportAddressHelper;
use Chill\MainBundle\Export\ListInterface;
use Chill\MainBundle\Form\Type\ChillDateType;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Chill\PersonBundle\Entity\Person;
use Chill\PersonBundle\Export\Declarations;
use Chill\PersonBundle\Export\Helper\ListPersonHelper;
use Chill\PersonBundle\Security\Authorization\PersonVoter;
use DateTimeImmutable;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\Query;
use PhpOffice\PhpSpreadsheet\Shared\Date;
use Symfony\Component\Form\Extension\Core\Type\ChoiceType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Validator\Constraints\Callback;
use Symfony\Component\Validator\Context\ExecutionContextInterface;

use function addcslashes;
use function array_key_exists;
use function array_keys;
use function array_merge;
use function count;
use function in_array;
use function strlen;
use function uniqid;

/**
 * Render a list of people.
 */
class ListPerson implements ExportElementValidatedInterface, ListInterface, GroupedExportInterface
{
    private $slugs = [];

    public function __construct(private readonly ExportAddressHelper $addressHelper, private readonly CustomFieldProvider $customFieldProvider, private readonly ListPersonHelper $listPersonHelper, private readonly EntityManagerInterface $entityManager, private readonly TranslatableStringHelper $translatableStringHelper)
    {
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $choices = array_combine(ListPersonHelper::FIELDS, ListPersonHelper::FIELDS);

        foreach ($this->getCustomFields() as $cf) {
            $choices[$this->translatableStringHelper->localize($cf->getName())]
                =
                $cf->getSlug();
        }

        // Add a checkbox to select fields
        $builder->add('fields', ChoiceType::class, [
            'multiple' => true,
            'expanded' => true,
            'choices' => $choices,
            'label' => 'Fields to include in export',
            'choice_attr' => static function (string $val): array {
                // add a 'data-display-target' for address fields
                if (str_starts_with($val, 'address') || 'center' === $val || 'household' === $val) {
                    return ['data-display-target' => 'address_date'];
                }

                return [];
            },
            'constraints' => [new Callback([
                'callback' => static function ($selected, ExecutionContextInterface $context) {
                    if (count($selected) === 0) {
                        $context->buildViolation('You must select at least one element')
                            ->atPath('fields')
                            ->addViolation();
                    }
                },
            ])],
            'data' => array_values($choices),
        ]);

        // add a date  field for addresses
        $builder->add('address_date', ChillDateType::class, [
            'label' => 'Data valid at this date',
            'help' => 'Data regarding center, addresses, and so on will be computed at this date',
            'data' => new DateTimeImmutable(),
            'input' => 'datetime_immutable',
        ]);
    }

    public function getAllowedFormattersTypes()
    {
        return [FormatterInterface::TYPE_LIST];
    }

    public function getDescription()
    {
        return 'Create a list of people according to various filters.';
    }

    public function getGroup(): string
    {
        return 'Exports of persons';
    }

    public function getLabels($key, array $values, $data)
    {
        if (in_array($key, $this->listPersonHelper->getAllPossibleFields(), true)) {
            return $this->listPersonHelper->getLabels($key, $values, $data);
        }

        return $this->getLabelForCustomField($key, $values, $data);
    }

    public function getQueryKeys($data)
    {
        $fields = [];

        foreach (ListPersonHelper::FIELDS as $key) {
            if (!in_array($key, $data['fields'], true)) {
                continue;
            }

            if (str_starts_with($key, 'address_fields')) {
                $fields = array_merge($fields, $this->addressHelper->getKeys(ExportAddressHelper::F_ALL, 'address_fields'));

                continue;
            }

            if ('lifecycleUpdate' === $key) {
                $fields = array_merge($fields, ['createdAt', 'createdBy', 'updatedAt', 'updatedBy']);

                continue;
            }

            $fields[] = $key;
        }

        // add the key from slugs and return
        return array_merge($fields, array_keys($this->slugs));
    }

    public function getResult($query, $data)
    {
        return $query->getQuery()->getResult(Query::HYDRATE_SCALAR);
    }

    public function getTitle()
    {
        return 'List peoples';
    }

    public function getType()
    {
        return Declarations::PERSON_TYPE;
    }

    /**
     * param array{fields: string[], address_date: DateTimeImmutable} $data.
     */
    public function initiateQuery(array $requiredModifiers, array $acl, array $data = [])
    {
        $centers = array_map(static fn ($el) => $el['center'], $acl);

        // throw an error if any fields are present
        if (!array_key_exists('fields', $data)) {
            throw new \Doctrine\DBAL\Exception\InvalidArgumentException('any fields '
                . 'have been checked');
        }

        $qb = $this->entityManager->createQueryBuilder();

        $qb
            ->from(Person::class, 'person')
            ->andWhere(
                $qb->expr()->exists(
                    'SELECT 1 FROM ' . Person\PersonCenterHistory::class . ' pch WHERE pch.person = person.id AND pch.center IN (:authorized_centers)'
                )
            )
            ->setParameter('authorized_centers', $centers);

        $fields = $data['fields'];

        $this->listPersonHelper->addSelect($qb, $fields, $data['address_date']);

        foreach ($this->getCustomFields() as $cf) {
            if (!in_array($cf->getSlug(), $fields, true)) {
                continue;
            }

            $cfType = $this->customFieldProvider->getCustomFieldByType($cf->getType());

            if ($cfType instanceof CustomFieldChoice && $cfType->isMultiple($cf)) {
                foreach ($cfType->getChoices($cf) as $choiceSlug => $label) {
                    $slug = $this->slugToDQL($cf->getSlug(), 'choice', ['choiceSlug' => $choiceSlug]);
                    $qb->addSelect(
                        sprintf(
                            'GET_JSON_FIELD_BY_KEY(person.cFData, :slug%s) AS %s',
                            $slug,
                            $slug
                        )
                    );
                    $qb->setParameter(sprintf('slug%s', $slug), $cf->getSlug());
                }
            } else {
                $slug = $this->slugToDQL($cf->getSlug());
                $qb->addSelect(
                    sprintf(
                        'GET_JSON_FIELD_BY_KEY(person.cFData, :slug%s) AS %s',
                        $slug,
                        $slug
                    )
                );
                $qb->setParameter(sprintf('slug%s', $slug), $cf->getSlug());
            }
        }

        return $qb;
    }

    public function requiredRole(): string
    {
        return PersonVoter::LISTS;
    }

    public function supportsModifiers()
    {
        return [Declarations::PERSON_TYPE, Declarations::PERSON_IMPLIED_IN];
    }

    public function validateForm($data, ExecutionContextInterface $context)
    {
        // get the field starting with address_
        $addressFields = array_filter(
            ListPersonHelper::FIELDS,
            static fn (string $el): bool => str_starts_with($el, 'address_')
        );

        // check if there is one field starting with address in data
        if (count(array_intersect($data['fields'], $addressFields)) > 0) {
            // if a field address is checked, the date must not be empty
            if (!$data['address_date'] instanceof DateTimeImmutable) {
                $context
                    ->buildViolation('You must set this date if an address is checked')
                    ->atPath('address_date')
                    ->addViolation();
            }
        }
    }

    private function DQLToSlug($cleanedSlug)
    {
        return $this->slugs[$cleanedSlug]['slug'];
    }

    /**
     * @return array An array with keys = 'slug', 'type', 'additionnalInfo'
     */
    private function extractInfosFromSlug(mixed $slug): array
    {
        return $this->slugs[$slug];
    }

    /**
     * Get custom fields associated with person.
     *
     * @return CustomField[]
     */
    private function getCustomFields()
    {
        return $this->entityManager
            ->createQuery('SELECT cf '
                . 'FROM ChillCustomFieldsBundle:CustomField cf '
                . 'JOIN cf.customFieldGroup g '
                . 'WHERE cf.type != :title AND g.entity LIKE :entity')
            ->setParameters([
                'title' => 'title',
                'entity' => addcslashes(Person::class, '\\'),
            ])
            ->getResult();
    }

    private function getLabelForCustomField($key, array $values, $data)
    {
        // for fields which are custom fields
        /** @var CustomField $cf */
        $cf = $this->entityManager
            ->getRepository(CustomField::class)
            ->findOneBy(['slug' => $this->DQLToSlug($key)]);
        $cfType = $this->customFieldProvider->getCustomFieldByType($cf->getType());
        $defaultFunction = function ($value) use ($cf) {
            if ('_header' === $value) {
                return $this->translatableStringHelper->localize($cf->getName());
            }

            if (null === $value) {
                return '';
            }

            return $this->customFieldProvider
                ->getCustomFieldByType($cf->getType())
                ->render(json_decode($value, true, 512, JSON_THROW_ON_ERROR), $cf, 'csv');
        };

        if ($cfType instanceof CustomFieldChoice && $cfType->isMultiple($cf)) {
            return function ($value) use ($cf, $cfType, $key) {
                $slugChoice = $this->extractInfosFromSlug($key)['additionnalInfos']['choiceSlug'];

                if (null === $value) {
                    return '';
                }
                $decoded = json_decode($value, true, 512, JSON_THROW_ON_ERROR);

                if ('_header' === $value) {
                    $label = $cfType->getChoices($cf)[$slugChoice];

                    return $this->translatableStringHelper->localize($cf->getName())
                            . ' | ' . $label;
                }

                if ('_other' === $slugChoice && $cfType->isChecked($cf, $slugChoice, $decoded)) {
                    return $cfType->extractOtherValue($cf, $decoded);
                }

                return $cfType->isChecked($cf, $slugChoice, $decoded);
            };
        }

        return $defaultFunction;
    }

    /**
     * Clean a slug to be usable by DQL.
     */
    private function slugToDQL(string $slug, string $type = 'default', array $additionalInfos = []): string
    {
        $uid = 'slug_' . uniqid('', false);

        $this->slugs[$uid] = [
            'slug' => $slug,
            'type' => $type,
            'additionnalInfos' => $additionalInfos,
        ];

        return $uid;
    }
}
