<?php

declare(strict_types=1);

/*
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

namespace Chill\PersonBundle\Export\Export;

use Chill\MainBundle\Export\ExportElementValidatedInterface;
use Chill\MainBundle\Export\FormatterInterface;
use Chill\MainBundle\Export\GroupedExportInterface;
use Chill\MainBundle\Export\Helper\ExportAddressHelper;
use Chill\MainBundle\Export\ListInterface;
use Chill\MainBundle\Form\Type\ChillDateType;
use Chill\PersonBundle\Entity\AccompanyingPeriod;
use Chill\PersonBundle\Entity\Person;
use Chill\PersonBundle\Entity\Person\PersonCenterHistory;
use Chill\PersonBundle\Export\Declarations;
use Chill\PersonBundle\Export\Helper\ListPersonHelper;
use Chill\PersonBundle\Security\Authorization\PersonVoter;
use DateTimeImmutable;
use Doctrine\ORM\AbstractQuery;
use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\Form\Extension\Core\Type\ChoiceType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Validator\Constraints\Callback;
use Symfony\Component\Validator\Context\ExecutionContextInterface;
use function array_key_exists;
use function count;
use function in_array;
use function strlen;

class ListPersonWithAccompanyingPeriod implements ExportElementValidatedInterface, ListInterface, GroupedExportInterface
{
    public function __construct(private readonly ExportAddressHelper $addressHelper, private readonly ListPersonHelper $listPersonHelper, private readonly EntityManagerInterface $entityManager)
    {
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $choices = array_combine(ListPersonHelper::FIELDS, ListPersonHelper::FIELDS);

        // Add a checkbox to select fields
        $builder->add('fields', ChoiceType::class, [
            'multiple' => true,
            'expanded' => true,
            'choices' => $choices,
            'label' => 'Fields to include in export',
            'choice_attr' => static function (string $val): array {
                // add a 'data-display-target' for address fields
                if (str_starts_with($val, 'address') || 'center' === $val || 'household' === $val) {
                    return ['data-display-target' => 'address_date'];
                }

                return [];
            },
            'constraints' => [new Callback([
                'callback' => static function ($selected, ExecutionContextInterface $context) {
                    if (count($selected) === 0) {
                        $context->buildViolation('You must select at least one element')
                            ->atPath('fields')
                            ->addViolation();
                    }
                },
            ])],
            'data' => array_values($choices),
        ]);

        // add a date  field for addresses
        $builder->add('address_date', ChillDateType::class, [
            'label' => 'Data valid at this date',
            'help' => 'Data regarding center, addresses, and so on will be computed at this date',
            'data' => new DateTimeImmutable(),
            'input' => 'datetime_immutable',
        ]);
    }

    public function getAllowedFormattersTypes()
    {
        return [FormatterInterface::TYPE_LIST];
    }

    public function getDescription()
    {
        return 'export.list.person_with_acp.Create a list of people having an accompaying periods, according to various filters.';
    }

    public function getGroup(): string
    {
        return 'Exports of persons';
    }

    public function getLabels($key, array $values, $data)
    {
        return $this->listPersonHelper->getLabels($key, $values, $data);
    }

    public function getQueryKeys($data)
    {
        $fields = [];

        foreach (ListPersonHelper::FIELDS as $key) {
            if (!in_array($key, $data['fields'], true)) {
                continue;
            }

            if (str_starts_with($key, 'address_fields')) {
                $fields = array_merge($fields, $this->addressHelper->getKeys(ExportAddressHelper::F_ALL, 'address_fields'));

                continue;
            }

            if ('lifecycleUpdate' === $key) {
                $fields = array_merge($fields, ['createdAt', 'createdBy', 'updatedAt', 'updatedBy']);

                continue;
            }

            $fields[] = $key;
        }

        return $fields;
    }

    public function getResult($query, $data)
    {
        return $query->getQuery()->getResult(AbstractQuery::HYDRATE_SCALAR);
    }

    public function getTitle()
    {
        return 'export.list.person_with_acp.List peoples having an accompanying period';
    }

    public function getType()
    {
        return Declarations::PERSON_TYPE;
    }

    /**
     * param array{fields: string[], address_date: DateTimeImmutable} $data.
     */
    public function initiateQuery(array $requiredModifiers, array $acl, array $data = [])
    {
        $centers = array_map(static fn ($el) => $el['center'], $acl);

        // throw an error if any fields are present
        if (!array_key_exists('fields', $data)) {
            throw new \Doctrine\DBAL\Exception\InvalidArgumentException('any fields '
                . 'have been checked');
        }

        $qb = $this->entityManager->createQueryBuilder();

        $qb->from(Person::class, 'person')
            ->join('person.accompanyingPeriodParticipations', 'acppart')
            ->join('acppart.accompanyingPeriod', 'acp')
            ->andWhere($qb->expr()->neq('acp.step', "'" . AccompanyingPeriod::STEP_DRAFT . "'"))
            ->andWhere(
                $qb->expr()->exists(
                    'SELECT 1 FROM ' . PersonCenterHistory::class . ' pch WHERE pch.person = person.id AND pch.center IN (:authorized_centers)'
                )
            )->setParameter('authorized_centers', $centers);

        $fields = $data['fields'];

        $this->listPersonHelper->addSelect($qb, $fields, $data['address_date']);

        return $qb;
    }

    public function requiredRole(): string
    {
        return PersonVoter::LISTS;
    }

    public function supportsModifiers()
    {
        return [Declarations::PERSON_TYPE, Declarations::PERSON_IMPLIED_IN, Declarations::ACP_TYPE];
    }

    public function validateForm($data, ExecutionContextInterface $context)
    {
        // get the field starting with address_
        $addressFields = array_filter(
            ListPersonHelper::FIELDS,
            static fn (string $el): bool => str_starts_with($el, 'address_')
        );

        // check if there is one field starting with address in data
        if (count(array_intersect($data['fields'], $addressFields)) > 0) {
            // if a field address is checked, the date must not be empty
            if (!$data['address_date'] instanceof DateTimeImmutable) {
                $context
                    ->buildViolation('You must set this date if an address is checked')
                    ->atPath('address_date')
                    ->addViolation();
            }
        }
    }
}
