<?php

declare(strict_types=1);

/*
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

namespace Chill\PersonBundle\Export\Filter\AccompanyingCourseFilters;

use Chill\MainBundle\Export\FilterInterface;
use Chill\MainBundle\Form\Type\PickRollingDateType;
use Chill\MainBundle\Service\RollingDate\RollingDate;
use Chill\MainBundle\Service\RollingDate\RollingDateConverterInterface;
use Chill\PersonBundle\Entity\AccompanyingPeriod;
use Chill\PersonBundle\Export\Declarations;
use Doctrine\ORM\QueryBuilder;
use Symfony\Component\Form\Extension\Core\Type\ChoiceType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Contracts\Translation\TranslatorInterface;
use function in_array;

class StepFilter implements FilterInterface
{
    private const A = 'acp_filter_bystep_stephistories';

    private const DEFAULT_CHOICE = AccompanyingPeriod::STEP_CONFIRMED;

    private const P = 'acp_step_filter_date';

    private const STEPS = [
        'Draft' => AccompanyingPeriod::STEP_DRAFT,
        'Confirmed' => AccompanyingPeriod::STEP_CONFIRMED,
        'Closed' => AccompanyingPeriod::STEP_CLOSED,
    ];

    public function __construct(private readonly RollingDateConverterInterface $rollingDateConverter, private readonly TranslatorInterface $translator)
    {
    }

    public function addRole(): ?string
    {
        return null;
    }

    public function alterQuery(QueryBuilder $qb, $data)
    {
        if (!in_array(self::A, $qb->getAllAliases(), true)) {
            $qb->leftJoin('acp.stepHistories', self::A);
        }

        $qb
            ->andWhere(
                $qb->expr()->andX(
                    $qb->expr()->lte(self::A . '.startDate', ':' . self::P),
                    $qb->expr()->orX(
                        $qb->expr()->isNull(self::A . '.endDate'),
                        $qb->expr()->lt(self::A . '.endDate', ':' . self::P)
                    )
                )
            )
            ->andWhere(
                $qb->expr()->in(self::A . '.step', ':acp_filter_by_step_steps')
            )
            ->setParameter(self::P, $this->rollingDateConverter->convert($data['calc_date']))
            ->setParameter('acp_filter_by_step_steps', $data['accepted_steps']);
    }

    public function applyOn()
    {
        return Declarations::ACP_TYPE;
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $builder
            ->add('accepted_steps', ChoiceType::class, [
                'choices' => self::STEPS,
                'multiple' => false,
                'expanded' => true,
                'empty_data' => self::DEFAULT_CHOICE,
                'data' => self::DEFAULT_CHOICE,
            ])
            ->add('calc_date', PickRollingDateType::class, [
                'label' => 'export.acp.filter.by_step.date_calc',
                'data' => new RollingDate(RollingDate::T_TODAY),
            ]);
    }

    public function describeAction($data, $format = 'string')
    {
        $step = array_flip(self::STEPS)[$data['accepted_steps']];

        return ['Filtered by steps: only %step%', [
            '%step%' => $this->translator->trans($step),
        ]];
    }

    public function getTitle()
    {
        return 'Filter by step';
    }
}
