<?php

declare(strict_types=1);

/*
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

namespace Chill\PersonBundle\Export\Filter\AccompanyingCourseFilters;

use Chill\MainBundle\Entity\UserJob;
use Chill\MainBundle\Export\FilterInterface;
use Chill\MainBundle\Form\Type\PickRollingDateType;
use Chill\MainBundle\Repository\UserJobRepositoryInterface;
use Chill\MainBundle\Service\RollingDate\RollingDate;
use Chill\MainBundle\Service\RollingDate\RollingDateConverterInterface;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Chill\PersonBundle\Export\Declarations;
use Doctrine\ORM\QueryBuilder;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Security\Core\Security;

class UserJobFilter implements FilterInterface
{
    private const A = 'acp_ujob_filter_uhistory';

    private const AU = 'acp_ujob_filter_uhistory_user';

    private const P = 'acp_ujob_filter_date';

    private const PJ = 'acp_ujob_filter_job';

    public function __construct(private readonly TranslatableStringHelper $translatableStringHelper, private readonly UserJobRepositoryInterface $userJobRepository, private readonly RollingDateConverterInterface $rollingDateConverter)
    {
    }

    public function addRole(): ?string
    {
        return null;
    }

    public function alterQuery(QueryBuilder $qb, $data)
    {
        $qb
            ->join('acp.userHistories', self::A)
            ->andWhere(
                $qb->expr()->andX(
                    $qb->expr()->lte(self::A . '.startDate', ':' . self::P),
                    $qb->expr()->orX(
                        $qb->expr()->isNull(self::A . '.endDate'),
                        $qb->expr()->gt(self::A . '.endDate', ':' . self::P)
                    )
                )
            )
            ->setParameter(
                self::P,
                $this->rollingDateConverter->convert($data['date_calc'])
            )
            ->join(self::A . '.user', self::AU)
            ->andWhere(
                $qb->expr()->in(self::AU . '.userJob', ':' . self::PJ)
            )
            ->setParameter(self::PJ, $data['jobs']);
    }

    public function applyOn()
    {
        return Declarations::ACP_TYPE;
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $builder
            ->add('jobs', EntityType::class, [
                'class' => UserJob::class,
                'choices' => $this->userJobRepository->findAllActive(),
                'multiple' => true,
                'expanded' => true,
                'choice_label' => fn (UserJob $job) => $this->translatableStringHelper->localize($job->getLabel()),
                'label' => 'Job',
            ])
            ->add('date_calc', PickRollingDateType::class, [
                'data' => new RollingDate(RollingDate::T_TODAY),
                'label' => 'export.filter.course.by_user_scope.Computation date for referrer',
                'required' => true,
            ]);
    }

    public function describeAction($data, $format = 'string')
    {
        return [
            'Filtered by user job: only %job%', [
                '%job%' => implode(
                    ', ',
                    array_map(
                        fn (UserJob $job) => $this->translatableStringHelper->localize($job->getLabel()),
                        $data['jobs']->toArray()
                    )
                ),
            ],
        ];
    }

    public function getTitle()
    {
        return 'Filter by user job';
    }
}
