<?php

declare(strict_types=1);

/*
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

namespace Chill\PersonBundle\Export\Filter\AccompanyingCourseFilters;

use Chill\MainBundle\Entity\Scope;
use Chill\MainBundle\Entity\User;
use Chill\MainBundle\Export\FilterInterface;
use Chill\MainBundle\Form\Type\PickRollingDateType;
use Chill\MainBundle\Repository\ScopeRepositoryInterface;
use Chill\MainBundle\Service\RollingDate\RollingDate;
use Chill\MainBundle\Service\RollingDate\RollingDateConverterInterface;
use Chill\MainBundle\Templating\TranslatableStringHelper;
use Chill\PersonBundle\Export\Declarations;
use Doctrine\ORM\QueryBuilder;
use Symfony\Bridge\Doctrine\Form\Type\EntityType;
use Symfony\Component\Form\FormBuilderInterface;
use Symfony\Component\Security\Core\Security;

class UserScopeFilter implements FilterInterface
{
    private const A = 'acp_uscope_filter_uhistory';

    private const AU = 'acp_uscope_filter_uhistory_user';

    private const P = 'acp_uscope_filter_date';

    private const PS = 'acp_uscope_filter_scopes';

    public function __construct(private readonly ScopeRepositoryInterface $scopeRepository, private readonly Security $security, private readonly TranslatableStringHelper $translatableStringHelper, private readonly RollingDateConverterInterface $rollingDateConverter)
    {
    }

    public function addRole(): ?string
    {
        return null;
    }

    public function alterQuery(QueryBuilder $qb, $data)
    {
        $qb
            ->join('acp.userHistories', self::A)
            ->andWhere(
                $qb->expr()->andX(
                    $qb->expr()->lte(self::A . '.startDate', ':' . self::P),
                    $qb->expr()->orX(
                        $qb->expr()->isNull(self::A . '.endDate'),
                        $qb->expr()->gt(self::A . '.endDate', ':' . self::P)
                    )
                )
            )
            ->setParameter(
                self::P,
                $this->rollingDateConverter->convert($data['date_calc'])
            )
            ->join(self::A . '.user', self::AU)
            ->andWhere(
                $qb->expr()->in(self::AU . '.mainScope', ':' . self::PS)
            )
            ->setParameter(self::PS, $data['scopes']);
    }

    public function applyOn()
    {
        return Declarations::ACP_TYPE;
    }

    public function buildForm(FormBuilderInterface $builder)
    {
        $builder
            ->add('scopes', EntityType::class, [
                'class' => Scope::class,
                'choices' => $this->scopeRepository->findAllActive(),
                'choice_label' => fn (Scope $s) => $this->translatableStringHelper->localize($s->getName()),
                'multiple' => true,
                'expanded' => true,
            ])
            ->add('date_calc', PickRollingDateType::class, [
                'data' => new RollingDate(RollingDate::T_TODAY),
                'label' => 'export.filter.course.by_user_scope.Computation date for referrer',
                'required' => true,
            ]);
    }

    public function describeAction($data, $format = 'string')
    {
        return [
            'Filtered by user main scope: only %scope%', [
                '%scope%' => implode(
                    ', ',
                    array_map(
                        fn (Scope $s) => $this->translatableStringHelper->localize($s->getName()),
                        $data['scopes']->toArray()
                    )
                ),
            ],
        ];
    }

    public function getTitle()
    {
        return 'Filter by user scope';
    }

    private function getUserMainScope(): Scope
    {
        /** @var User $user */
        $user = $this->security->getUser();

        return $user->getMainScope();
    }
}
