<?php

declare(strict_types=1);

/*
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

namespace Chill\PersonBundle\Repository;

use Chill\MainBundle\Entity\Address;
use Chill\MainBundle\Entity\Location;
use Chill\MainBundle\Entity\PostalCode;
use Chill\MainBundle\Entity\Scope;
use Chill\MainBundle\Entity\User;
use Chill\MainBundle\Entity\UserJob;
use Chill\MainBundle\Security\Authorization\AuthorizationHelper;
use Chill\MainBundle\Security\Resolver\CenterResolverManagerInterface;
use Chill\PersonBundle\Entity\AccompanyingPeriod;
use Chill\PersonBundle\Entity\AccompanyingPeriodParticipation;
use Chill\PersonBundle\Entity\Household\PersonHouseholdAddress;
use Chill\PersonBundle\Entity\Person;
use Chill\PersonBundle\Security\Authorization\AccompanyingPeriodVoter;
use DateTime;

use DateTimeImmutable;
use Doctrine\DBAL\Types\Types;
use Doctrine\ORM\Query\Expr\Join;
use Doctrine\ORM\QueryBuilder;
use Symfony\Component\Security\Core\Security;
use function count;

final readonly class AccompanyingPeriodACLAwareRepository implements AccompanyingPeriodACLAwareRepositoryInterface
{
    public function __construct(
        private AccompanyingPeriodRepository $accompanyingPeriodRepository,
        private Security $security,
        private AuthorizationHelper $authorizationHelper,
        private CenterResolverManagerInterface $centerResolverDispatcher
    ) {
    }

    /**
     * @param array<PostalCode> $postalCodes
     */
    public function buildQueryOpenedAccompanyingCourseByUser(?User $user, array $postalCodes = []): QueryBuilder
    {
        $qb = $this->accompanyingPeriodRepository->createQueryBuilder('ap');

        $qb->where($qb->expr()->eq('ap.user', ':user'))
            ->andWhere(
                $qb->expr()->neq('ap.step', ':draft'),
                $qb->expr()->orX(
                    $qb->expr()->isNull('ap.closingDate'),
                    $qb->expr()->gt('ap.closingDate', ':now')
                )
            )
            ->setParameter('user', $user)
            ->setParameter('now', new DateTime('now'))
            ->setParameter('draft', AccompanyingPeriod::STEP_DRAFT);

        if ([] !== $postalCodes) {
            $qb->join('ap.locationHistories', 'location_history')
                ->leftJoin(PersonHouseholdAddress::class, 'person_address', Join::WITH, 'IDENTITY(location_history.personLocation) = IDENTITY(person_address.person)')
                ->join(
                    Address::class,
                    'address',
                    Join::WITH,
                    'COALESCE(IDENTITY(location_history.addressLocation), IDENTITY(person_address.address)) = address.id'
                )
                ->andWhere(
                    $qb->expr()->orX(
                        $qb->expr()->isNull('person_address'),
                        $qb->expr()->andX(
                            $qb->expr()->lte('person_address.validFrom', ':now'),
                            $qb->expr()->orX(
                                $qb->expr()->isNull('person_address.validTo'),
                                $qb->expr()->lt('person_address.validTo', ':now')
                            )
                        )
                    )
                )
                ->andWhere(
                    $qb->expr()->isNull('location_history.endDate')
                )
                ->andWhere(
                    $qb->expr()->in('address.postcode', ':postal_codes')
                )
                ->setParameter('now', new DateTimeImmutable('now'), Types::DATE_IMMUTABLE)
                ->setParameter('postal_codes', $postalCodes);
        }

        return $qb;
    }

    public function countByUnDispatched(array $jobs, array $services, array $administrativeLocations): int
    {
        $qb = $this->addACLByUnDispatched($this->buildQueryUnDispatched($jobs, $services, $administrativeLocations));

        $qb->select('COUNT(ap)');

        return $qb->getQuery()->getSingleScalarResult();
    }

    public function countByUserAndPostalCodesOpenedAccompanyingPeriod(?User $user, array $postalCodes): int
    {
        if (null === $user) {
            return 0;
        }

        return $this->buildQueryOpenedAccompanyingCourseByUser($user, $postalCodes)
            ->select('COUNT(ap)')
            ->getQuery()
            ->getSingleScalarResult();
    }

    public function countByUserOpenedAccompanyingPeriod(?User $user): int
    {
        if (null === $user) {
            return 0;
        }

        return $this->buildQueryOpenedAccompanyingCourseByUser($user)
            ->select('COUNT(ap)')
            ->getQuery()
            ->getSingleScalarResult();
    }

    public function findByPerson(
        Person $person,
        string $role,
        ?array $orderBy = [],
        ?int $limit = null,
        ?int $offset = null
    ): array {
        $qb = $this->accompanyingPeriodRepository->createQueryBuilder('ap');
        $scopes = $this->authorizationHelper
            ->getReachableCircles(
                $this->security->getUser(),
                $role,
                $this->centerResolverDispatcher->resolveCenters($person)
            );

        if (0 === count($scopes)) {
            return [];
        }

        $qb
            ->join('ap.participations', 'participation')
            ->where($qb->expr()->eq('participation.person', ':person'))
            ->andWhere(
                $qb->expr()->orX(
                    'ap.confidential = FALSE',
                    $qb->expr()->eq('ap.user', ':user')
                )
            )
            ->andWhere(
                $qb->expr()->orX(
                    $qb->expr()->neq('ap.step', ':draft'),
                    $qb->expr()->orX(
                        $qb->expr()->eq('ap.createdBy', ':creator'),
                        $qb->expr()->isNull('ap.createdBy')
                    )
                )
            )
            ->setParameter('draft', AccompanyingPeriod::STEP_DRAFT)
            ->setParameter('person', $person)
            ->setParameter('user', $this->security->getUser())
            ->setParameter('creator', $this->security->getUser());
        // add join condition for scopes
        $orx = $qb->expr()->orX(
            $qb->expr()->eq('ap.step', ':draft')
        );

        foreach ($scopes as $key => $scope) {
            $orx->add($qb->expr()->orX(
                $qb->expr()->isMemberOf(':scope_' . $key, 'ap.scopes'),
                $qb->expr()->eq('ap.user', ':user')
            ));
            $qb->setParameter('scope_' . $key, $scope);
            $qb->setParameter('user', $this->security->getUser());
        }
        $qb->andWhere($orx);

        return $qb->getQuery()->getResult();
    }

    public function findByUnDispatched(array $jobs, array $services, array $administrativeLocations, ?int $limit = null, ?int $offset = null): array
    {
        $qb = $this->addACLByUnDispatched($this->buildQueryUnDispatched($jobs, $services, $administrativeLocations));

        $qb->select('ap');

        if (null !== $limit) {
            $qb->setMaxResults($limit);
        }

        if (null !== $offset) {
            $qb->setFirstResult($offset);
        }

        return $qb->getQuery()->getResult();
    }

    public function findByUserAndPostalCodesOpenedAccompanyingPeriod(?User $user, array $postalCodes, array $orderBy = [], int $limit = 0, int $offset = 50): array
    {
        if (null === $user) {
            return [];
        }

        $qb = $this->buildQueryOpenedAccompanyingCourseByUser($user);

        $qb->setFirstResult($offset)
            ->setMaxResults($limit);

        foreach ($orderBy as $field => $direction) {
            $qb->addOrderBy('ap.' . $field, $direction);
        }

        return $qb->getQuery()->getResult();
    }

    /**
     * @return array|AccompanyingPeriod[]
     */
    public function findByUserOpenedAccompanyingPeriod(?User $user, array $orderBy = [], int $limit = 0, int $offset = 50): array
    {
        if (null === $user) {
            return [];
        }

        $qb = $this->buildQueryOpenedAccompanyingCourseByUser($user);

        $qb->setFirstResult($offset)
            ->setMaxResults($limit);

        foreach ($orderBy as $field => $direction) {
            $qb->addOrderBy('ap.' . $field, $direction);
        }

        return $qb->getQuery()->getResult();
    }

    private function addACLByUnDispatched(QueryBuilder $qb): QueryBuilder
    {
        $centers = $this->authorizationHelper->getReachableCenters(
            $this->security->getUser(),
            AccompanyingPeriodVoter::SEE
        );

        $orX = $qb->expr()->orX();

        if (0 === count($centers)) {
            return $qb->andWhere("'FALSE' = 'TRUE'");
        }

        foreach ($centers as $key => $center) {
            $scopes = $this->authorizationHelper
                ->getReachableCircles(
                    $this->security->getUser(),
                    AccompanyingPeriodVoter::SEE,
                    $center
                );

            $and = $qb->expr()->andX(
                $qb->expr()->exists('SELECT part FROM ' . AccompanyingPeriodParticipation::class . ' part ' .
                    "JOIN part.person p WHERE part.accompanyingPeriod = ap.id AND p.center = :center_{$key}")
            );
            $qb->setParameter('center_' . $key, $center);
            $orScope = $qb->expr()->orX();

            foreach ($scopes as $skey => $scope) {
                $orScope->add(
                    $qb->expr()->isMemberOf(':scope_' . $key . '_' . $skey, 'ap.scopes')
                );
                $qb->setParameter('scope_' . $key . '_' . $skey, $scope);
            }

            $and->add($orScope);
            $orX->add($and);
        }

        return $qb->andWhere($orX);
    }

    /**
     * @param array|UserJob[] $jobs
     * @param array|Scope[] $services
     * @param array|Location[] $locations
     */
    private function buildQueryUnDispatched(array $jobs, array $services, array $locations): QueryBuilder
    {
        $qb = $this->accompanyingPeriodRepository->createQueryBuilder('ap');

        $qb->where(
            $qb->expr()->andX(
                $qb->expr()->isNull('ap.user'),
                $qb->expr()->neq('ap.step', ':draft'),
                $qb->expr()->neq('ap.step', ':closed')
            )
        )
            ->setParameter('draft', AccompanyingPeriod::STEP_DRAFT)
            ->setParameter('closed', AccompanyingPeriod::STEP_CLOSED);

        if (0 < count($jobs)) {
            $qb->andWhere($qb->expr()->in('ap.job', ':jobs'))
                ->setParameter('jobs', $jobs);
        }

        if (0 < count($locations)) {
            $qb->andWhere($qb->expr()->in('ap.administrativeLocation', ':locations'))
                ->setParameter('locations', $locations);
        }

        if (0 < count($services)) {
            $or = $qb->expr()->orX();

            foreach ($services as $key => $service) {
                $or->add($qb->expr()->isMemberOf(':scope_' . $key, 'ap.scopes'));
                $qb->setParameter('scope_' . $key, $service);
            }
            $qb->andWhere($or);
        }

        return $qb;
    }
}
